/* 
 * Copyright (C) 2015-2016 Pier Luigi Fiorini
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GREENISLAND_SCREENCASTER_CLIENT_PROTOCOL_H
#define GREENISLAND_SCREENCASTER_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct greenisland_screencaster;
struct greenisland_screencast;

extern const struct wl_interface greenisland_screencaster_interface;
extern const struct wl_interface greenisland_screencast_interface;

#ifndef GREENISLAND_SCREENCASTER_ERROR_ENUM
#define GREENISLAND_SCREENCASTER_ERROR_ENUM
/**
 * greenisland_screencaster_error - greenisland_screencaster error values
 * @GREENISLAND_SCREENCASTER_ERROR_BAD_OUTPUT: The client passed an
 *	invalid output
 *
 * These errors can be emitted in response to greenisland_screencaster
 * requests.
 */
enum greenisland_screencaster_error {
	GREENISLAND_SCREENCASTER_ERROR_BAD_OUTPUT = 0,
};
#endif /* GREENISLAND_SCREENCASTER_ERROR_ENUM */

#define GREENISLAND_SCREENCASTER_CAPTURE	0

static inline void
greenisland_screencaster_set_user_data(struct greenisland_screencaster *greenisland_screencaster, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) greenisland_screencaster, user_data);
}

static inline void *
greenisland_screencaster_get_user_data(struct greenisland_screencaster *greenisland_screencaster)
{
	return wl_proxy_get_user_data((struct wl_proxy *) greenisland_screencaster);
}

static inline void
greenisland_screencaster_destroy(struct greenisland_screencaster *greenisland_screencaster)
{
	wl_proxy_destroy((struct wl_proxy *) greenisland_screencaster);
}

static inline struct greenisland_screencast *
greenisland_screencaster_capture(struct greenisland_screencaster *greenisland_screencaster, struct wl_output *output)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) greenisland_screencaster,
			 GREENISLAND_SCREENCASTER_CAPTURE, &greenisland_screencast_interface, NULL, output);

	return (struct greenisland_screencast *) id;
}

#ifndef GREENISLAND_SCREENCAST_ERROR_ENUM
#define GREENISLAND_SCREENCAST_ERROR_ENUM
/**
 * greenisland_screencast_error - greenisland_screencast error values
 * @GREENISLAND_SCREENCAST_ERROR_BAD_BUFFER: Buffer doesn't match
 *	requirements
 *
 * These errors are emitted with the 'failed' event when a capture
 * request fails.
 */
enum greenisland_screencast_error {
	GREENISLAND_SCREENCAST_ERROR_BAD_BUFFER = 0,
};
#endif /* GREENISLAND_SCREENCAST_ERROR_ENUM */

#ifndef GREENISLAND_SCREENCAST_TRANSFORM_ENUM
#define GREENISLAND_SCREENCAST_TRANSFORM_ENUM
/**
 * greenisland_screencast_transform - frame buffer transform
 * @GREENISLAND_SCREENCAST_TRANSFORM_NORMAL: No particular transform has
 *	been applied
 * @GREENISLAND_SCREENCAST_TRANSFORM_Y_INVERTED: Y axis is inverted
 *
 * These values are emitted with the 'frame' event and represent the
 * transform that is applied to the frame buffer.
 */
enum greenisland_screencast_transform {
	GREENISLAND_SCREENCAST_TRANSFORM_NORMAL = 1,
	GREENISLAND_SCREENCAST_TRANSFORM_Y_INVERTED = 2,
};
#endif /* GREENISLAND_SCREENCAST_TRANSFORM_ENUM */

/**
 * greenisland_screencast - frame capture
 * @setup: notify buffer size requirements
 * @frame: notify a frame was recorded, or an error
 * @failed: capture failed
 * @cancelled: notify a request was cancelled
 *
 * This interface drives the frame capture into client-allocated buffers.
 * Immediately after this object is created, the compositor sends the
 * 'setup' event with the buffer size. The client uses this information to
 * allocate a buffer to hold the frame image.
 *
 * When the buffer is created, the 'record' request must be invoked each
 * time the client wants to record a frame.
 */
struct greenisland_screencast_listener {
	/**
	 * setup - notify buffer size requirements
	 * @width: (none)
	 * @height: (none)
	 * @stride: (none)
	 * @format: (none)
	 *
	 * This event will be sent immediately after creating the
	 * greenisland_screencast object.
	 *
	 * The wl_buffer passed to the 'record' request must be big enough
	 * to accomodate an image with the given width, height and format.
	 * If it's smaller, the compositor will send the 'failed' event.
	 *
	 * The buffer must be a shm buffer, trying to use another type of
	 * buffer will result in a failure. The format argument value is
	 * defined in the wl_shm::format enum.
	 *
	 * This event might be sent again later in the life-time of the
	 * object, should the pending capture requests are canceled.
	 */
	void (*setup)(void *data,
		      struct greenisland_screencast *greenisland_screencast,
		      int32_t width,
		      int32_t height,
		      int32_t stride,
		      int32_t format);
	/**
	 * frame - notify a frame was recorded, or an error
	 * @buffer: client buffer that holds the data
	 * @time: when the frame was recorded
	 * @transform: optional transform applied to the image
	 *
	 * The compositor will send this event after a frame was
	 * recorded, or in case an error has happened.
	 *
	 * The 'time' argument is the time the compositor recorded the
	 * frame, in milliseconds.
	 *
	 * Clients must call 'record' again to record the next frame.
	 */
	void (*frame)(void *data,
		      struct greenisland_screencast *greenisland_screencast,
		      struct wl_buffer *buffer,
		      uint32_t time,
		      int32_t transform);
	/**
	 * failed - capture failed
	 * @error: error that has occurred
	 * @buffer: frame buffer
	 *
	 * The value of the error argument is one of the 'error' enum.
	 */
	void (*failed)(void *data,
		       struct greenisland_screencast *greenisland_screencast,
		       int32_t error,
		       struct wl_buffer *buffer);
	/**
	 * cancelled - notify a request was cancelled
	 * @buffer: old frame buffer
	 *
	 * This event is sent if the client calls 'record' more than once
	 * for the same frame. The frame will be recorded using a new
	 * buffer. The 'buffer' argument is the old buffer.
	 */
	void (*cancelled)(void *data,
			  struct greenisland_screencast *greenisland_screencast,
			  struct wl_buffer *buffer);
};

static inline int
greenisland_screencast_add_listener(struct greenisland_screencast *greenisland_screencast,
				    const struct greenisland_screencast_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) greenisland_screencast,
				     (void (**)(void)) listener, data);
}

#define GREENISLAND_SCREENCAST_DESTROY	0
#define GREENISLAND_SCREENCAST_RECORD	1

static inline void
greenisland_screencast_set_user_data(struct greenisland_screencast *greenisland_screencast, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) greenisland_screencast, user_data);
}

static inline void *
greenisland_screencast_get_user_data(struct greenisland_screencast *greenisland_screencast)
{
	return wl_proxy_get_user_data((struct wl_proxy *) greenisland_screencast);
}

static inline void
greenisland_screencast_destroy(struct greenisland_screencast *greenisland_screencast)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_screencast,
			 GREENISLAND_SCREENCAST_DESTROY);

	wl_proxy_destroy((struct wl_proxy *) greenisland_screencast);
}

static inline void
greenisland_screencast_record(struct greenisland_screencast *greenisland_screencast, struct wl_buffer *buffer)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_screencast,
			 GREENISLAND_SCREENCAST_RECORD, buffer);
}

#ifdef  __cplusplus
}
#endif

#endif
