/* 
 * Copyright (C) 2015-2016 Pier Luigi Fiorini
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GREENISLAND_SCREENSHOOTER_CLIENT_PROTOCOL_H
#define GREENISLAND_SCREENSHOOTER_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct greenisland_screenshooter;
struct greenisland_screenshot;

extern const struct wl_interface greenisland_screenshooter_interface;
extern const struct wl_interface greenisland_screenshot_interface;

#ifndef GREENISLAND_SCREENSHOOTER_EFFECTS_ENUM
#define GREENISLAND_SCREENSHOOTER_EFFECTS_ENUM
/**
 * greenisland_screenshooter_effects - effects of the screenshot
 * @GREENISLAND_SCREENSHOOTER_EFFECTS_POINTER: include the pointer
 * @GREENISLAND_SCREENSHOOTER_EFFECTS_BORDER: include the window border
 *
 * The different effects for the screenshot.
 *
 * Desktop environments may extend this enum by taking up a range of values
 * and documenting the range they chose in this description. They are not
 * required to document the values for the range that they chose. Ideally,
 * any good extensions from a desktop environment should make its way into
 * standardization into this enum.
 *
 * The current reserved ranges are:
 *
 * 0x0000 - 0x0FFF: greenisland-screenshooter core values, documented
 * below.
 */
enum greenisland_screenshooter_effects {
	GREENISLAND_SCREENSHOOTER_EFFECTS_POINTER = 1,
	GREENISLAND_SCREENSHOOTER_EFFECTS_BORDER = 2,
};
#endif /* GREENISLAND_SCREENSHOOTER_EFFECTS_ENUM */

#define GREENISLAND_SCREENSHOOTER_CAPTURE_OUTPUT	0
#define GREENISLAND_SCREENSHOOTER_CAPTURE_ACTIVE	1
#define GREENISLAND_SCREENSHOOTER_CAPTURE_SURFACE	2
#define GREENISLAND_SCREENSHOOTER_CAPTURE_AREA	3

static inline void
greenisland_screenshooter_set_user_data(struct greenisland_screenshooter *greenisland_screenshooter, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) greenisland_screenshooter, user_data);
}

static inline void *
greenisland_screenshooter_get_user_data(struct greenisland_screenshooter *greenisland_screenshooter)
{
	return wl_proxy_get_user_data((struct wl_proxy *) greenisland_screenshooter);
}

static inline void
greenisland_screenshooter_destroy(struct greenisland_screenshooter *greenisland_screenshooter)
{
	wl_proxy_destroy((struct wl_proxy *) greenisland_screenshooter);
}

static inline struct greenisland_screenshot *
greenisland_screenshooter_capture_output(struct greenisland_screenshooter *greenisland_screenshooter, struct wl_output *output, struct wl_array *effects)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) greenisland_screenshooter,
			 GREENISLAND_SCREENSHOOTER_CAPTURE_OUTPUT, &greenisland_screenshot_interface, NULL, output, effects);

	return (struct greenisland_screenshot *) id;
}

static inline struct greenisland_screenshot *
greenisland_screenshooter_capture_active(struct greenisland_screenshooter *greenisland_screenshooter, struct wl_array *effects)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) greenisland_screenshooter,
			 GREENISLAND_SCREENSHOOTER_CAPTURE_ACTIVE, &greenisland_screenshot_interface, NULL, effects);

	return (struct greenisland_screenshot *) id;
}

static inline struct greenisland_screenshot *
greenisland_screenshooter_capture_surface(struct greenisland_screenshooter *greenisland_screenshooter, struct wl_array *effects)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) greenisland_screenshooter,
			 GREENISLAND_SCREENSHOOTER_CAPTURE_SURFACE, &greenisland_screenshot_interface, NULL, effects);

	return (struct greenisland_screenshot *) id;
}

static inline struct greenisland_screenshot *
greenisland_screenshooter_capture_area(struct greenisland_screenshooter *greenisland_screenshooter, struct wl_array *effects)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) greenisland_screenshooter,
			 GREENISLAND_SCREENSHOOTER_CAPTURE_AREA, &greenisland_screenshot_interface, NULL, effects);

	return (struct greenisland_screenshot *) id;
}

#ifndef GREENISLAND_SCREENSHOT_ERROR_ENUM
#define GREENISLAND_SCREENSHOT_ERROR_ENUM
/**
 * greenisland_screenshot_error - greenisland_screenshot error values
 * @GREENISLAND_SCREENSHOT_ERROR_BAD_BUFFER: Buffer doesn't match
 *	requirements
 *
 * These errors are emitted with the 'failed' event when a capture
 * request fails.
 */
enum greenisland_screenshot_error {
	GREENISLAND_SCREENSHOT_ERROR_BAD_BUFFER = 0,
};
#endif /* GREENISLAND_SCREENSHOT_ERROR_ENUM */

/**
 * greenisland_screenshot - screenshot
 * @setup: notify buffer size requirements
 * @done: emitted when the buffer is filled
 * @failed: capture failed
 *
 * A screenshot allows a client to perform a capture of an entire output,
 * an area selected by a user-driven operation, the active window or a
 * specific surface. Whatever is to be captured is decided by the capture
 * request that was called from the greenisland_screenshooter interface.
 *
 * The 'setup' event is sent as soon as the client binds to this interface
 * and indicates the size of the buffer that needs to be allocated.
 *
 * Invoking the 'capture' request triggers the actual capture. The 'done'
 * event is sent when the capture is done, at this point the client can
 * decide what to do with the buffer (store into a file, display to the
 * user, ...).
 */
struct greenisland_screenshot_listener {
	/**
	 * setup - notify buffer size requirements
	 * @width: width of the frame, in pixels
	 * @height: height of the frame, in pixels
	 * @stride: stride of the buffer
	 * @format: format of the buffer
	 *
	 * This event will be sent immediately after creating the
	 * greenisland_screenshot object.
	 *
	 * The wl_buffer passed to the 'record' request must be big enough
	 * to accomodate an image with the given width, height and format.
	 * If it's smaller, the compositor will send the 'failed' event.
	 *
	 * The buffer must be a shm buffer, trying to use another type of
	 * buffer will result in a failure. The format argument value is
	 * defined in the wl_shm::format enum.
	 */
	void (*setup)(void *data,
		      struct greenisland_screenshot *greenisland_screenshot,
		      int32_t width,
		      int32_t height,
		      int32_t stride,
		      int32_t format);
	/**
	 * done - emitted when the buffer is filled
	 *
	 * The done event is sent when the buffer is filled. This object
	 * can be destroyed calling 'destroy' afterwards.
	 */
	void (*done)(void *data,
		     struct greenisland_screenshot *greenisland_screenshot);
	/**
	 * failed - capture failed
	 * @error: error that has occurred
	 *
	 * The value of the error argument is one if the 'error' enum.
	 */
	void (*failed)(void *data,
		       struct greenisland_screenshot *greenisland_screenshot,
		       int32_t error);
};

static inline int
greenisland_screenshot_add_listener(struct greenisland_screenshot *greenisland_screenshot,
				    const struct greenisland_screenshot_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) greenisland_screenshot,
				     (void (**)(void)) listener, data);
}

#define GREENISLAND_SCREENSHOT_DESTROY	0
#define GREENISLAND_SCREENSHOT_RECORD	1

static inline void
greenisland_screenshot_set_user_data(struct greenisland_screenshot *greenisland_screenshot, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) greenisland_screenshot, user_data);
}

static inline void *
greenisland_screenshot_get_user_data(struct greenisland_screenshot *greenisland_screenshot)
{
	return wl_proxy_get_user_data((struct wl_proxy *) greenisland_screenshot);
}

static inline void
greenisland_screenshot_destroy(struct greenisland_screenshot *greenisland_screenshot)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_screenshot,
			 GREENISLAND_SCREENSHOT_DESTROY);

	wl_proxy_destroy((struct wl_proxy *) greenisland_screenshot);
}

static inline void
greenisland_screenshot_record(struct greenisland_screenshot *greenisland_screenshot, struct wl_buffer *buffer)
{
	wl_proxy_marshal((struct wl_proxy *) greenisland_screenshot,
			 GREENISLAND_SCREENSHOT_RECORD, buffer);
}

#ifdef  __cplusplus
}
#endif

#endif
