/* 
 * Copyright (C) 2015-2016 Pier Luigi Fiorini
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GREENISLAND_SCREENSHOOTER_SERVER_PROTOCOL_H
#define GREENISLAND_SCREENSHOOTER_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-server.h"

struct wl_client;
struct wl_resource;

struct greenisland_screenshooter;
struct greenisland_screenshot;

extern const struct wl_interface greenisland_screenshooter_interface;
extern const struct wl_interface greenisland_screenshot_interface;

#ifndef GREENISLAND_SCREENSHOOTER_EFFECTS_ENUM
#define GREENISLAND_SCREENSHOOTER_EFFECTS_ENUM
/**
 * greenisland_screenshooter_effects - effects of the screenshot
 * @GREENISLAND_SCREENSHOOTER_EFFECTS_POINTER: include the pointer
 * @GREENISLAND_SCREENSHOOTER_EFFECTS_BORDER: include the window border
 *
 * The different effects for the screenshot.
 *
 * Desktop environments may extend this enum by taking up a range of values
 * and documenting the range they chose in this description. They are not
 * required to document the values for the range that they chose. Ideally,
 * any good extensions from a desktop environment should make its way into
 * standardization into this enum.
 *
 * The current reserved ranges are:
 *
 * 0x0000 - 0x0FFF: greenisland-screenshooter core values, documented
 * below.
 */
enum greenisland_screenshooter_effects {
	GREENISLAND_SCREENSHOOTER_EFFECTS_POINTER = 1,
	GREENISLAND_SCREENSHOOTER_EFFECTS_BORDER = 2,
};
#endif /* GREENISLAND_SCREENSHOOTER_EFFECTS_ENUM */

/**
 * greenisland_screenshooter - dump the contents of an output, a surface
 *	or an area to a buffer
 * @capture_output: take a screenshot of an entire output
 * @capture_active: take a screenshot of the active window
 * @capture_surface: take a screenshot of a surface
 * @capture_area: take a screenshot of an area
 *
 * This interface is used by programs to take a screenshot of an entire
 * output, a user-selected area or a specific surface.
 *
 * The client asks the compositor to capture with one of the capture
 * requests. As a result, a greenisland_screenshot object will be created
 * and the greenisland_screenshot.setup event will be sent as soon as the
 * client binds to the interface.
 *
 * See the documentation of greenisland_screencast for more details.
 */
struct greenisland_screenshooter_interface {
	/**
	 * capture_output - take a screenshot of an entire output
	 * @id: (none)
	 * @output: (none)
	 * @effects: effects to apply for this capture
	 *
	 * Asks the compositor to capture an entire output.
	 */
	void (*capture_output)(struct wl_client *client,
			       struct wl_resource *resource,
			       uint32_t id,
			       struct wl_resource *output,
			       struct wl_array *effects);
	/**
	 * capture_active - take a screenshot of the active window
	 * @id: (none)
	 * @effects: effects to apply for this capture
	 *
	 * Asks the compositor to take a screenshot of the active window.
	 */
	void (*capture_active)(struct wl_client *client,
			       struct wl_resource *resource,
			       uint32_t id,
			       struct wl_array *effects);
	/**
	 * capture_surface - take a screenshot of a surface
	 * @id: (none)
	 * @effects: effects to apply for this capture
	 *
	 * Asks the compositor to take a screenshot of a surface of a
	 * window selected by the user.
	 */
	void (*capture_surface)(struct wl_client *client,
				struct wl_resource *resource,
				uint32_t id,
				struct wl_array *effects);
	/**
	 * capture_area - take a screenshot of an area
	 * @id: (none)
	 * @effects: effects to apply for this capture
	 *
	 * Asks the compositor to take a screenshot of an area. This
	 * request starts a user-driven selection of the area that will be
	 * captured.
	 */
	void (*capture_area)(struct wl_client *client,
			     struct wl_resource *resource,
			     uint32_t id,
			     struct wl_array *effects);
};


#ifndef GREENISLAND_SCREENSHOT_ERROR_ENUM
#define GREENISLAND_SCREENSHOT_ERROR_ENUM
/**
 * greenisland_screenshot_error - greenisland_screenshot error values
 * @GREENISLAND_SCREENSHOT_ERROR_BAD_BUFFER: Buffer doesn't match
 *	requirements
 *
 * These errors are emitted with the 'failed' event when a capture
 * request fails.
 */
enum greenisland_screenshot_error {
	GREENISLAND_SCREENSHOT_ERROR_BAD_BUFFER = 0,
};
#endif /* GREENISLAND_SCREENSHOT_ERROR_ENUM */

/**
 * greenisland_screenshot - screenshot
 * @destroy: destroy greenisland_screenshot
 * @record: take a screenshot
 *
 * A screenshot allows a client to perform a capture of an entire output,
 * an area selected by a user-driven operation, the active window or a
 * specific surface. Whatever is to be captured is decided by the capture
 * request that was called from the greenisland_screenshooter interface.
 *
 * The 'setup' event is sent as soon as the client binds to this interface
 * and indicates the size of the buffer that needs to be allocated.
 *
 * Invoking the 'capture' request triggers the actual capture. The 'done'
 * event is sent when the capture is done, at this point the client can
 * decide what to do with the buffer (store into a file, display to the
 * user, ...).
 */
struct greenisland_screenshot_interface {
	/**
	 * destroy - destroy greenisland_screenshot
	 *
	 * Destroy this greenisland_screenshot object.
	 */
	void (*destroy)(struct wl_client *client,
			struct wl_resource *resource);
	/**
	 * record - take a screenshot
	 * @buffer: client buffer that will hold data
	 *
	 * Ask the compositor to take a screenshot and save the contents
	 * into the buffer. What is being capture can be either an output,
	 * a surface or an area that was selected with user interaction,
	 * depending on the greenisland_screenshooter request that was
	 * invoked.
	 */
	void (*record)(struct wl_client *client,
		       struct wl_resource *resource,
		       struct wl_resource *buffer);
};

#define GREENISLAND_SCREENSHOT_SETUP	0
#define GREENISLAND_SCREENSHOT_DONE	1
#define GREENISLAND_SCREENSHOT_FAILED	2

#define GREENISLAND_SCREENSHOT_SETUP_SINCE_VERSION	1
#define GREENISLAND_SCREENSHOT_DONE_SINCE_VERSION	1
#define GREENISLAND_SCREENSHOT_FAILED_SINCE_VERSION	1

static inline void
greenisland_screenshot_send_setup(struct wl_resource *resource_, int32_t width, int32_t height, int32_t stride, int32_t format)
{
	wl_resource_post_event(resource_, GREENISLAND_SCREENSHOT_SETUP, width, height, stride, format);
}

static inline void
greenisland_screenshot_send_done(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, GREENISLAND_SCREENSHOT_DONE);
}

static inline void
greenisland_screenshot_send_failed(struct wl_resource *resource_, int32_t error)
{
	wl_resource_post_event(resource_, GREENISLAND_SCREENSHOT_FAILED, error);
}

#ifdef  __cplusplus
}
#endif

#endif
