/* 
 * Copyright (C) 2015-2016 Pier Luigi Fiorini
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef GREENISLAND_SERVER_PROTOCOL_H
#define GREENISLAND_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-server.h"

struct wl_client;
struct wl_resource;

struct greenisland_applications;
struct greenisland_windows;
struct greenisland_window;

extern const struct wl_interface greenisland_applications_interface;
extern const struct wl_interface greenisland_windows_interface;
extern const struct wl_interface greenisland_window_interface;

/**
 * greenisland_applications - track running applications
 * @quit: quit an application
 *
 * This interface is used by UI modules interested in tracking running
 * applications, for example a library that expose the list of running
 * applications to a launcher. Only one client can bind this interface at a
 * time.
 */
struct greenisland_applications_interface {
	/**
	 * quit - quit an application
	 * @app_id: application identifier
	 *
	 * Ask the compositor to terminate a Wayland client with the
	 * specified application identifier.
	 */
	void (*quit)(struct wl_client *client,
		     struct wl_resource *resource,
		     const char *app_id);
};

#define GREENISLAND_APPLICATIONS_REGISTERED	0
#define GREENISLAND_APPLICATIONS_UNREGISTERED	1
#define GREENISLAND_APPLICATIONS_FOCUSED	2
#define GREENISLAND_APPLICATIONS_UNFOCUSED	3

#define GREENISLAND_APPLICATIONS_REGISTERED_SINCE_VERSION	1
#define GREENISLAND_APPLICATIONS_UNREGISTERED_SINCE_VERSION	1
#define GREENISLAND_APPLICATIONS_FOCUSED_SINCE_VERSION	1
#define GREENISLAND_APPLICATIONS_UNFOCUSED_SINCE_VERSION	1

static inline void
greenisland_applications_send_registered(struct wl_resource *resource_, const char *app_id, uint32_t pid)
{
	wl_resource_post_event(resource_, GREENISLAND_APPLICATIONS_REGISTERED, app_id, pid);
}

static inline void
greenisland_applications_send_unregistered(struct wl_resource *resource_, const char *app_id, uint32_t pid)
{
	wl_resource_post_event(resource_, GREENISLAND_APPLICATIONS_UNREGISTERED, app_id, pid);
}

static inline void
greenisland_applications_send_focused(struct wl_resource *resource_, const char *app_id)
{
	wl_resource_post_event(resource_, GREENISLAND_APPLICATIONS_FOCUSED, app_id);
}

static inline void
greenisland_applications_send_unfocused(struct wl_resource *resource_, const char *app_id)
{
	wl_resource_post_event(resource_, GREENISLAND_APPLICATIONS_UNFOCUSED, app_id);
}

#ifndef GREENISLAND_WINDOWS_TYPE_ENUM
#define GREENISLAND_WINDOWS_TYPE_ENUM
enum greenisland_windows_type {
	GREENISLAND_WINDOWS_TYPE_TOPLEVEL = 1,
	GREENISLAND_WINDOWS_TYPE_POPUP = 2,
	GREENISLAND_WINDOWS_TYPE_TRANSIENT = 3,
};
#endif /* GREENISLAND_WINDOWS_TYPE_ENUM */

#ifndef GREENISLAND_WINDOWS_STATE_ENUM
#define GREENISLAND_WINDOWS_STATE_ENUM
enum greenisland_windows_state {
	GREENISLAND_WINDOWS_STATE_INACTIVE = 0x0,
	GREENISLAND_WINDOWS_STATE_ACTIVE = 0x1,
	GREENISLAND_WINDOWS_STATE_MINIMIZED = 0x2,
	GREENISLAND_WINDOWS_STATE_MAXIMIZED = 0x3,
	GREENISLAND_WINDOWS_STATE_FULLSCREEN = 0x4,
	GREENISLAND_WINDOWS_STATE_KEEP_ABOVE = 0x5,
	GREENISLAND_WINDOWS_STATE_KEEP_BELOW = 0x6,
	GREENISLAND_WINDOWS_STATE_SHADED = 0x7,
	GREENISLAND_WINDOWS_STATE_ON_ALL_WORKSPACES = 0x8,
};
#endif /* GREENISLAND_WINDOWS_STATE_ENUM */

#define GREENISLAND_WINDOWS_WINDOW_MAPPED	0

#define GREENISLAND_WINDOWS_WINDOW_MAPPED_SINCE_VERSION	1

static inline void
greenisland_windows_send_window_mapped(struct wl_resource *resource_, struct wl_resource *id, uint32_t type, uint32_t state, const char *title, const char *app_id)
{
	wl_resource_post_event(resource_, GREENISLAND_WINDOWS_WINDOW_MAPPED, id, type, state, title, app_id);
}

#define GREENISLAND_WINDOW_TITLE_CHANGED	0
#define GREENISLAND_WINDOW_APP_ID_CHANGED	1
#define GREENISLAND_WINDOW_STATE_CHANGED	2
#define GREENISLAND_WINDOW_WORKSPACE_CHANGED	3
#define GREENISLAND_WINDOW_UNMAPPED	4

#define GREENISLAND_WINDOW_TITLE_CHANGED_SINCE_VERSION	1
#define GREENISLAND_WINDOW_APP_ID_CHANGED_SINCE_VERSION	1
#define GREENISLAND_WINDOW_STATE_CHANGED_SINCE_VERSION	1
#define GREENISLAND_WINDOW_WORKSPACE_CHANGED_SINCE_VERSION	1
#define GREENISLAND_WINDOW_UNMAPPED_SINCE_VERSION	1

static inline void
greenisland_window_send_title_changed(struct wl_resource *resource_, const char *title)
{
	wl_resource_post_event(resource_, GREENISLAND_WINDOW_TITLE_CHANGED, title);
}

static inline void
greenisland_window_send_app_id_changed(struct wl_resource *resource_, const char *app_id)
{
	wl_resource_post_event(resource_, GREENISLAND_WINDOW_APP_ID_CHANGED, app_id);
}

static inline void
greenisland_window_send_state_changed(struct wl_resource *resource_, uint32_t flags)
{
	wl_resource_post_event(resource_, GREENISLAND_WINDOW_STATE_CHANGED, flags);
}

static inline void
greenisland_window_send_workspace_changed(struct wl_resource *resource_, uint32_t number)
{
	wl_resource_post_event(resource_, GREENISLAND_WINDOW_WORKSPACE_CHANGED, number);
}

static inline void
greenisland_window_send_unmapped(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, GREENISLAND_WINDOW_UNMAPPED);
}

#ifdef  __cplusplus
}
#endif

#endif
